/* Energy Components, http://www.energy-components.com
 * Tietoenator AS,    http://www.tietoenator.com
 *
 * Copyright (c) 2005 Tietoenator AS.  All rights reserved.
 * --------------------------------------------------------
 * $Header: /product/com.ec.frmw.jboss.service/src/com/ec/frmw/jboss/service/ECSchedulerService.java,v 1.3 2009/05/22 12:35:10 grimschr Exp $
 */
package com.ec.frmw.jboss.service;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.management.ManagementFactory;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.InetAddress;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServerConnection;
import javax.management.MBeanServerFactory;
import javax.management.MalformedObjectNameException;
import javax.management.Notification;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.quartz.InterruptableJob;
import org.quartz.Job;
import org.quartz.JobExecutionContext;
import org.quartz.JobKey;
import org.quartz.Scheduler;
import org.quartz.SchedulerException;
import org.quartz.SchedulerFactory;
import org.quartz.SchedulerMetaData;
import org.quartz.UnableToInterruptJobException;
import org.quartz.impl.SchedulerRepository;
import org.quartz.impl.StdSchedulerFactory;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * ECSchedulerService is a MBean responsible for starting, stopping and
 * monitoring all EC Quartz instances running on this server.
 * 
 * @author xxmbuamu
 */
public class ECSchedulerService implements ECSchedulerServiceMBean,
		I_FileWatchTimerCallback {
	private static JMXConnector connector;

	private static MBeanServerConnection connection;

	final Logger log = Logger.getLogger(ECSchedulerService.class);

	// The partition/cluster name
	String partitionName = System.getProperty("jboss.partition.name",
			"DefaultPartition");

	// Host name
	String hostID = null;

	/*
	 * ------------------------------------------------ Constructors
	 */
	public ECSchedulerService() throws IOException {
		connect("127.0.0.1", "8080");
	}

	public String getName() {
		return org.jboss.util.Classes.stripPackageName(log.getName());
	}

	// NOTYET: This _should_ be used for the LogManager.setRepositorySelector
	// private static Object guard = new Object();

	/*
	 * ------------------------------------------------ Life Cycle Methods
	 */
	public void start() throws Exception {

		String mdcKey = ((Properties) ecQuartzProperties
				.get(GLOBAL_SERVER_NAME))
				.getProperty("org.quartz.plugin.Log4JLoggerPlugin.mdcKey");
		log.debug("Setting new ThreadRepositorySelector(" + mdcKey + ")");
		if (mdcKey != null) {
			LogManager.setRepositorySelector(new ThreadRepositorySelector(
					mdcKey), null);
		}

		log.info("MBean [" + "EnergyComponents:name=ECSchedulerService " + "] starting...");
		// Call super
		// super.startService();

		// get the host name
//		ObjectName objectName = new ObjectName("jboss:service=Naming");
//		Object oValue = server.getAttribute(objectName, "Port");
		ObjectName objectName = new ObjectName("jboss.as:subsystem=naming");
		Object oValue = connection.getAttribute(objectName, "Port");
		hostID = InetAddress.getLocalHost().getHostName() + "." + oValue;

		// Check that we have a valid configuration.
		if (ecQuartzProperties.isEmpty()) {
			throw new Exception("Service [" + "EnergyComponents:name=ECSchedulerService"
					+ "] can't start. Invalid configuration.");
		}

		// Start up the Quartz Servers
		try {
			startQuartzServers(true);
		} catch (SchedulerException e) {
			log.warn("Problem when starting Quartz Servers: " + e.getMessage());
		} catch (Exception e) {
			log.info("NOTE: We are unable to start any Quartz servers at this time."
					+ " The Energy Components application is probably not deplyed yet. The service "
					+ "EnergyComponents:name=ECSchedulerService"
					+ " will still be running and starting the Quartz servers when Energy Components is deployed."
					+ " Message: [" + e.getMessage() + "]");
		}

		// subscribe for notification, true=dynamic
//		super.subscribe(true); // listener is me!

		// Call get instance to make sure the timer is started.
		FileWatchTimerTask.getInstance(); // this will start the timer if it not
											// already running

		// registerRPCHandler
		
		//Todo:Ticket sent to redpill on how to implement this.
		/**
		log.debug("registerRPCHandler");
		ObjectName obj = new ObjectName("jboss:service=" + partitionName); // NOTYET:
																			// What
																			// if
																			// the
																			// cluster
																			// is
																			// just
																			// not
																			// loaded
																			// yet?
																			// Can
																			// we
																			// depend
																			// on
																			// it?
		try {
			HAPartition partition = (HAPartition) this.server.getAttribute(obj,
					"HAPartition");
			partition.registerRPCHandler(serviceName.toString(), this);
		} catch (InstanceNotFoundException e) {
			log.info("HAPartition not found, we are running in a non-clustered JBoss environment");
		}
		**/

		log.info("MBean [" + "EnergyComponents:name=ECSchedulerService" + "] started successfully.");
	}

	public void stop() throws Exception {
		log.info("MBean [" + "EnergyComponents:name=ECSchedulerService" + "] stopping...");

		FileWatchTimerTask.removeFileSubscription(this);

		//Todo: Ticket sent to redpill asking how to implement this.
		// unRegisterRPCHandler
		/**
		log.debug("unRegisterRPCHandler");
		ObjectName obj = new ObjectName("jboss:service=" + partitionName);
		try {
			HAPartition partition = (HAPartition) this.server.getAttribute(obj,
					"HAPartition");
			partition.unregisterRPCHandler(serviceName.toString(), this);
		} catch (InstanceNotFoundException e) {
			// NO-OP
		}
**/

		stopQuartzServers();

		log.info("MBean [" + "EnergyComponents:name=ECSchedulerService" + "] stoped.");
	}

	/*
	 * ------------------------------------------------ Notification Methods
	 */
	public void handleNotification2(Notification notification, Object handback) {

		String notificationStr = "CLASS:" + notification.getClass() + ",TYPE:"
				+ notification.getType() + ",SOURCE:"
				+ notification.getSource() + ",MSG:"
				+ notification.getMessage();

		log.debug("Got notification. \n: [" + notificationStr + "]");

		String action = handback + "";
		if (!ecActionList.containsKey(action)) {
			// Nothing to do....
			log.warn("A notification arrived with a 'handback' that have no definition.");
			return;
		}

		Map filterList = (Map) ecActionList.get(action);
		log.debug("FilterList: " + filterList);
		for (Iterator i = filterList.keySet().iterator(); i.hasNext();) {
			String actionFilter = (String) i.next();
			String actionDo = (String) filterList.get(actionFilter);

			// log.debug("Matching: [" + actionFilter + "].");

			// Filter the notification
			if (notificationStr.matches(actionFilter)) {

				log.debug("Notification matched filter: [" + actionFilter
						+ "] on action [" + action + "]");
				log.debug("Executing action           : [" + actionDo + "].");

				// Do it!
				invokeAction(actionDo);

			} else {
				log.debug("Notification did not match filter: [" + actionFilter
						+ "] on action [" + action + "]");
			}
		}

	}

	@Override
	public void fileWatchTimerCallback(String file, String action) {
		// This is the callback method from the file Timer task

		invokeAction(action);

	}

	private void invokeAction(String actionDo) {
		log.info("Starting action: [" + actionDo + "]");
		try {
			if (STOP_ALL.equalsIgnoreCase(actionDo)) {
				stopQuartzServers();
			} else if (START_MISSING.equalsIgnoreCase(actionDo)) {
				startMissingQuartzServers();
			} else if (RESTART_ALL.equalsIgnoreCase(actionDo)) {
				startQuartzServers(true);
			} else if (STOP_ORPHAN.equalsIgnoreCase(actionDo)) {
				stopOrphanQuartzServers();
			} else if (STOP_ORPHAN_AND_START_MISSING.equalsIgnoreCase(actionDo)) {
				stopOrphanQuartzServers();
				startMissingQuartzServers();
			} else {
				log.warn("Unknown action: [" + actionDo + "].");
			}
		} catch (SchedulerException e) {
			log.warn("Problem when accessing the Quartz API: " + e.getMessage());
		} catch (Exception e) {
			log.warn("We are unable to start/stop any Quartz servers at this time beacause the the communication with Energy Components failed with message: ["
					+ e.getMessage() + "]");
		}
		log.info("...action [" + actionDo + "] done.");
	}

	public String getClusterPartitionName() {
		return partitionName;
	}

	public boolean isNodeMemberOfCluster() {

	boolean isMember = false;
/**
		try {
			ObjectName obj = new ObjectName("jboss:service="
					+ getClusterPartitionName());
			log.info("Looking up [" + obj.getCanonicalName() + "]...");
			HAPartition partition = (HAPartition) this.server.getAttribute(obj,
					"HAPartition");
			isMember = true;
			log.info("Looks like we are in a Cluster:");
			log.info("  Cluster/Partition:" + partition.getPartitionName());
			log.info("  CurrentNode      :" + partition.getNodeName());
			log.info("  List of Nodes    :" + partition.getCurrentView());
			log.info("  #Node in Cluster :" + partition.getCurrentView().size());

		} catch (InstanceNotFoundException e) {
			log.info("No Cluster found, we're running standalone.");
		} catch (MalformedObjectNameException e) {
			log.error("Problem querying JMX: " + e.getMessage());
		} catch (AttributeNotFoundException e) {
			log.error("Problem querying JMX: " + e.getMessage());
		} catch (MBeanException e) {
			log.error("Problem querying JMX: " + e.getMessage());
		} catch (ReflectionException e) {
			log.error("Problem querying JMX: " + e.getMessage());
		}
		**/
		return isMember;
	}

	protected boolean validateServerDBconnection(String name) throws Exception {
		String dbUser = null;
		Connection conn = null;
		PreparedStatement ps = null;
		ResultSet rs = null;
		int schedulecount = -1;

		log.debug("Validating DB...");
		try {
			// Try to get a connection...
			conn = (Connection) callExtObject(
					"com.ec.eccore.util.sql.ConnectionFactory", null, null,
					"getConnection", new Class[] { String.class },
					new Object[] { name });
			dbUser = conn.getMetaData().getUserName();

			ps = conn.prepareStatement(DB_SANITY_CHECK_SQL);
			rs = ps.executeQuery();
			if (rs.next()) {
				schedulecount = rs.getInt(1);
				log.info("Connection for ECSystem [" + name
						+ "] is valid, username is: [" + dbUser
						+ "] and SCHEDULE table has " + schedulecount
						+ " rows.");
			} else {
				log.warn("Oooops! Unable to find any rows in SCHEDULE table in ECSystem ["
						+ name
						+ "], dbUser is ["
						+ dbUser
						+ "]. Ignoring this database connection!");
			}
		} catch (SQLException e) {
			log.warn("Oooops! Can't find the SCHEDULE table in ECSystem ["
					+ name + "], dbUser is [" + dbUser
					+ "]. Ignoring this database connection because of: ["
					+ e.getMessage() + "]");
		} finally {
			// ...close the connection.
			Class[] argTypes2 = { Connection.class, Statement.class,
					ResultSet.class };
			Object[] argValues2 = { conn, ps, rs };
			callExtObject("com.ec.eccore.util.sql.ConnectionFactory", null,
					null, "cleanUp", argTypes2, argValues2);
		}

		return (schedulecount >= 0);
	}

	public void startQuartzServers(boolean restart) throws Exception {
		Exception exc = null;
		String ecSystem = null;
		Collection ecSystemNames = getEcSystems();

		if (ecSystemNames == null || ecSystemNames.size() == 0) {
			log.error("Cound not start any Quartz Server instances: There are no EC Systems defined, either no EC application are deployed or databaseconfig.xml has errors.");
		} else {
			for (Iterator i = ecSystemNames.iterator(); i.hasNext();) {
				try {
					ecSystem = (String) i.next();
					startQuartzServer(ecSystem, restart);
				} catch (Exception e) {
					exc = e;
					log.error("Problem starting Quartz Server [" + ecSystem
							+ "]: " + e.getMessage());
				}
			}
			if (exc != null) {
				throw exc;
			}
		}
	}

	public void startQuartzServer(String name, boolean restart)
			throws SchedulerException {

		Properties prop = getQuartzProperties(name);
		SchedulerFactory sf = new StdSchedulerFactory(prop);
		boolean dbValid = false;

		log.info("*** <> ***");
		log.info("*** [" + name + "] starting Quartz Scheduler instance...");
		if (!restart && sf.getScheduler(name) != null) {
			log.info("*** ["
					+ name
					+ "] Quartz Scheduler is already running, it will _not_ be restarted.");
			return;
		}

		// Stop exciting server
		if (restart && sf.getScheduler(name) != null) {
			stopQuartzServer(name);
		}

		// Ignore if we are marked as SHUTDOWN
		String state = prop.getProperty(PROP_KEY_STARTUP_STATE,
				SERVER_STATE_RUNNING).trim();
		if (SERVER_STATE_SHUTDOWN.equals(state)) {
			log.info("*** [" + name + "] Ignoring server, it's in [" + state
					+ "] state.");
			return;
		}

		// Do a sanity check on the DB....
		String StartWithoutSanityCheck = prop.getProperty(
				PROP_KEY_IGNORE_SANITY_CHECK, "FALSE").trim();

		if ("true".equalsIgnoreCase(StartWithoutSanityCheck)) {
			log.info("*** [" + name
					+ "] Sanity check is configured to be ignored.");
		} else {
			log.info("*** [" + name + "] Performing sanity check.");

			try {
				dbValid = validateServerDBconnection(name);
			} catch (Exception e) {
				log.info("*** ["
						+ name
						+ "] Unable to start server, validateServerDBconnection() failed: ["
						+ e.getMessage() + "]");
			}

			// Set state to SHUTDOWN if DB is not valid.
			if (!dbValid) {
				log.info("*** ["
						+ name
						+ "] Ignoring server, it did not pass DB validation and will be put into ["
						+ SERVER_STATE_SHUTDOWN + "] state.");
				return;
			}
		}

		// Ok, start the engines!
		log.info("*** [" + name + "] Starting Quartz Scheduler in state ["
				+ state + "]...");
		Scheduler scheduler = sf.getScheduler();
		if (SERVER_STATE_RUNNING.equals(state)) {
			scheduler.start();
		} else if (SERVER_STATE_STANDBY.equals(state)) {
			scheduler.standby();
		} else {
			log.warn("*** [" + name + "] Can't set scheduler state: [" + state
					+ "], unknown state.");
		}
		log.info("*** [" + name + "] Started Quartz Scheduler.");
	}

	public void startMissingQuartzServers() throws SecurityException,
			IllegalArgumentException, ExceptionInInitializerError,
			ClassNotFoundException, NoSuchMethodException,
			InstantiationException, IllegalAccessException,
			InvocationTargetException, SchedulerException {

		SchedulerFactory sf = new StdSchedulerFactory();
		Collection ecSystemNames = getEcSystems();
		SchedulerException exc = null;

		if (ecSystemNames == null || ecSystemNames.size() == 0) {
			log.error("Cound not start any Quartz Server instances: There are no EC Systems defined, either no EC application are deployed or databaseconfig.xml has errors.");
		} else {

			for (Iterator i = ecSystemNames.iterator(); i.hasNext();) {
				String ecSystem = (String) i.next();
				if (sf.getScheduler(ecSystem) == null) {
					log.info("Quartz Server [" + ecSystem
							+ "] is not running, starting it now...");
					try {
						startQuartzServer(ecSystem, true);
					} catch (Exception e) {
						exc = new SchedulerException(e);
						log.error("Problem starting Quartz Server [" + ecSystem
								+ "]: " + e.getMessage());
					}

				} else {
					log.info("Ignoring Quartz Server [" + ecSystem
							+ "], it is already running.");
				}
			}
			if (exc != null) {
				throw exc;
			}

		}
	}

	public void stopQuartzServers() {
		Scheduler quartzServer = null;
		for (Iterator i = SchedulerRepository.getInstance().lookupAll()
				.iterator(); i.hasNext();) {
			quartzServer = (Scheduler) i.next();
			stopQuartzServer(quartzServer);

			// NOTYET: BIG HACK! Need to call lookupAll() again to avoid
			// "java.util.ConcurrentModificationException" !
			i = SchedulerRepository.getInstance().lookupAll().iterator();
		}
	}

	public void stopOrphanQuartzServers() throws SecurityException,
			IllegalArgumentException, ExceptionInInitializerError,
			ClassNotFoundException, NoSuchMethodException,
			InstantiationException, IllegalAccessException,
			InvocationTargetException {
		Collection ecSystemNames = getEcSystems();
		Scheduler quartzServer = null;
		try {
			for (Iterator i = SchedulerRepository.getInstance().lookupAll()
					.iterator(); i.hasNext();) {
				quartzServer = (Scheduler) i.next();
				String quartzName = quartzServer.getSchedulerName();

				if (!ecSystemNames.contains(quartzName)) {
					log.info("Quartz Server [" + quartzName
							+ "] is no longer valid, stopping it.");
					stopQuartzServer(quartzServer);
				} else {
					log.info("Quartz Server [" + quartzName + "] is valid.");
				}
			}
		} catch (SchedulerException e) {
			log.error("Cound not stop Quartz Server instance [" + quartzServer
					+ "]: " + e.getMessage());
		}

	}

	public void stopQuartzServer(String name) throws SchedulerException {
		Scheduler sched = SchedulerRepository.getInstance().lookup(name);
		stopQuartzServer(sched);
	}

	public void stopQuartzServer(Scheduler scheduler) {
		String name = null;
		try {
			if (scheduler != null) {
				name = scheduler.getSchedulerName();
				boolean clean = "true".equalsIgnoreCase(("" + scheduler
						.getContext().getString("cleanShutdown")).trim());
				log.info("Stopping Quartz Scheduler instance: [" + name
						+ "], waitForJobsToCpmplete: [" + clean + "]...");
				scheduler.shutdown(clean);
				log.info("Stopped Quartz Scheduler instance: [" + name + "].");
			} else {
				log.warn("Can't stop a NULL Scheduler instance.");
			}
		} catch (SchedulerException e) {
			log.error("Cound not stop Quartz Server instance [" + name + "]: "
					+ e.getMessage());
		}

	}

	/*
	 * ------------------------------------------------ EC Methods
	 */
	public Collection getEcSystems() throws SecurityException,
			IllegalArgumentException, ExceptionInInitializerError,
			ClassNotFoundException, NoSuchMethodException,
			InstantiationException, IllegalAccessException,
			InvocationTargetException {
		// Try to call the ConnectionFactory in com.ec.eccore. It may not be
		// available, but give it a try.
		Collection ecSystems = null;
		ecSystems = (Collection) callExtObject(
				"com.ec.eccore.util.sql.ConnectionFactory", null, null,
				"getDatabaseNames", null, null);
		log.debug("The current set of EC systems is now: " + ecSystems);

		return ecSystems;
	}

	public void setSubscriptionAction(Element subscriptionActions)
			throws Exception {
		log.info("--Loading Actions--");
		ecActionList.clear();

		if (!XML_SUBSCRIPTION_ACTIONS.equals(subscriptionActions.getTagName())) {
			throw new Exception("Required element '" + XML_SUBSCRIPTION_ACTIONS
					+ " is missing in the 'SubscriptionAction' attribute.");
		}

		NodeList actionListElements = subscriptionActions
				.getElementsByTagName(XML_ACTION);
		for (int i = 0; i < actionListElements.getLength(); i++) {
			Element action = (Element) actionListElements.item(i);
			String actionName = action.getAttribute("name") + "";
			log.debug("Action [" + actionName + "] :");

			Map filterList = new HashMap();
			NodeList filterListElements = action
					.getElementsByTagName(XML_FILTER);
			for (int j = 0; j < filterListElements.getLength(); j++) {
				Element filter = (Element) filterListElements.item(j);
				String filterMatch = filter.getAttribute("match") + "";
				String filterDo = filter.getAttribute("do") + "";
				log.debug("  --> FilterAction [" + filterMatch + "]["
						+ filterDo + "]");
				filterList.put(filterMatch, filterDo);
			}
			ecActionList.put(actionName, filterList);
		}
		log.info("Loaded " + ecActionList.size() + " actions.");
	}

	public void setQuartzProperties(Element quartzProperties) throws Exception {
		log.info("--Loading Quartz Configurations--");
		ecQuartzProperties.clear();
		Properties prop = new Properties();
		InputStream is = null;

		if (!XML_QUARTZ_SERVER_CONFIG.equals(quartzProperties.getTagName())) {
			throw new Exception("Required element '" + XML_QUARTZ_SERVER_CONFIG
					+ " is missing in the 'QuartzProperties' attribute.");
		}

		Element serverEl = (Element) (quartzProperties
				.getElementsByTagName(XML_SERVER_GLOBAL).item(0));
		String serverGlobal = serverEl.getFirstChild().getNodeValue();

		is = new ByteArrayInputStream(serverGlobal.getBytes());
		prop.load(is);
		ecQuartzProperties.put(GLOBAL_SERVER_NAME, prop);
		log.info(" : [" + GLOBAL_SERVER_NAME + "] loaded with " + prop.size()
				+ " properties.");

		// Add initial startup state
		prop.put(PROP_KEY_INITIAL_STARTUP_STATE,
				prop.get(PROP_KEY_STARTUP_STATE));
		NodeList serverList = quartzProperties
				.getElementsByTagName(XML_SERVER_INSTANCE);
		for (int i = 0; i < serverList.getLength(); i++) {
			Element server = (Element) serverList.item(i);
			String serverName = server.getAttribute("name") + "";
			String serverProp = server.getFirstChild().getNodeValue();
			// Load the global properties
			// Then add the specific properties for this instance.
			Properties newProp = new Properties();
			is = new ByteArrayInputStream(serverGlobal.getBytes());
			newProp.load(is);

			is = new ByteArrayInputStream(serverProp.getBytes());
			newProp.load(is);
			ecQuartzProperties.put(serverName, newProp);

			// Add initial startup state
			newProp.put(PROP_KEY_INITIAL_STARTUP_STATE,
					newProp.get(PROP_KEY_STARTUP_STATE));
			log.info(" : [" + serverName + "] loaded with " + newProp.size()
					+ " properties.");
		}

		// DEBUG - start
		if (log.isDebugEnabled()) {
			for (Iterator i = ecQuartzProperties.keySet().iterator(); i
					.hasNext();) {

				String key1 = (String) i.next();
				Properties theProp = (Properties) ecQuartzProperties.get(key1);
				log.debug("--Configuration for [" + key1 + "]:");

				for (Iterator iter = theProp.keySet().iterator(); iter
						.hasNext();) {
					String key2 = (String) iter.next();
					log.debug("Key [" + key2 + "] = ["
							+ theProp.getProperty(key2) + "]");
				}

			}
		}
		// DEBUG - end

	}

	public void setFileSubscriptionList(Element fileSubscriptions)
			throws Exception {
		log.info("--Loading FileNotification--");

		if (!XML_FILE_NSL.equals(fileSubscriptions.getTagName())) {
			throw new Exception(
					"Required element '"
							+ XML_FILE_NSL
							+ " is missing in the 'FileNotificationSubscriptionList' attribute.");
		}

		long fileRefreshPeriod = Long.parseLong(fileSubscriptions
				.getAttribute("refreshPeriod"));
		FileWatchTimerTask.setFileRefresh(fileRefreshPeriod);

		NodeList actionListElements = fileSubscriptions
				.getElementsByTagName(XML_FILE);
		for (int i = 0; i < actionListElements.getLength(); i++) {
			Element action = (Element) actionListElements.item(i);
			String fileName = action.getAttribute("name") + "";
			String fileAction = action.getAttribute("do") + "";
			FileWatchTimerTask.addFileSubscription(fileName, fileAction, this);
			log.info("File Notification on [" + fileName + "], action: ["
					+ fileAction + "]");
		}
	}

	public String getECSchedulePinTo() {
		if (isNodeMemberOfCluster()) {
			return getClusterPartitionName();
		} else {
			return getLocalNodeName();
		}
	}

	public Properties getQuartzProperties(String serverName) {

		String jndiPrefix = "java:jdbc/"; // default value;
		try {
			jndiPrefix = (String) callExtObject(
					"com.ec.eccore.util.sql.ConnectionFactory", null, null,
					"getConfigAttribute", new Class[] { String.class,
							String.class }, new Object[] { serverName,
							"jndiroot" });
		} catch (Exception e) {
			log.error("Problem getting the jndiroot' (will now use default: "
					+ jndiPrefix + "):" + e.getMessage());
		}
		// log.debug("jndiRoot is :" + jndiPrefix);

		Properties prop = (Properties) ecQuartzProperties
				.get((ecQuartzProperties.containsKey(serverName) ? serverName
						: GLOBAL_SERVER_NAME));

		// There are some properties that always depend on the serverName.
		prop.setProperty("org.quartz.scheduler.instanceName", serverName);

		prop.setProperty("org.quartz.scheduler.instanceId", hostID + "."
				+ serverName);

		// Configure the DataSources
		prop.setProperty("org.quartz.jobStore.dataSource", serverName
				+ "_MANAGED");
		prop.setProperty("org.quartz.jobStore.driverDelegateClass",
				"org.quartz.impl.jdbcjobstore.oracle.OracleDelegate");
		prop.setProperty("org.quartz.jobStore.nonManagedTXDataSource",
				serverName + "_UN_MANAGED");

		// Get Connection from JNDI
		// prop.setProperty("org.quartz.dataSource." + serverName+ ".jndiURL",
		// jndiPrefix + serverName);

		// Get Connection the ECConnectionProvider
		prop.setProperty("org.quartz.dataSource." + serverName + "_MANAGED"
				+ ".connectionProvider.class",
				"com.ec.frmw.jboss.service.ECConnectionProvider");
		prop.setProperty("org.quartz.dataSource." + serverName + "_MANAGED"
				+ ".JNDIRoot", jndiPrefix);
		prop.setProperty("org.quartz.dataSource." + serverName + "_MANAGED"
				+ ".SystemName", serverName);
		prop.setProperty("org.quartz.dataSource." + serverName + "_MANAGED"
				+ ".ManagedConnection", "true");
		prop.setProperty("org.quartz.dataSource." + serverName + "_MANAGED"
				+ ".ECSchedulePinTo", getECSchedulePinTo());

		prop.setProperty("org.quartz.dataSource." + serverName + "_UN_MANAGED"
				+ ".connectionProvider.class",
				"com.ec.frmw.jboss.service.ECConnectionProvider");
		prop.setProperty("org.quartz.dataSource." + serverName + "_UN_MANAGED"
				+ ".JNDIRoot", jndiPrefix);
		prop.setProperty("org.quartz.dataSource." + serverName + "_UN_MANAGED"
				+ ".SystemName", serverName);
		prop.setProperty("org.quartz.dataSource." + serverName + "_UN_MANAGED"
				+ ".ManagedConnection", "false");
		prop.setProperty("org.quartz.dataSource." + serverName + "_UN_MANAGED"
				+ ".ECSchedulePinTo", getECSchedulePinTo());

		return prop;
	}

	protected List callClusterMethod(String methodName, Class[] argTypes,
			Object[] argValues) throws Exception {
		log.info("callClusterMethod(" + methodName + ", " + argValues + ", "
				+ argValues + ")");
		ArrayList result = null;

		//Todo: Ticket with Redpill about how to implement this
		ObjectName obj = new ObjectName("jboss:service=" + partitionName);
/**
		try {
			HAPartition partition = (HAPartition) this.server.getAttribute(obj,
					"HAPartition");
			log.info("About to call " + methodName + "(" + argValues
					+ ") on all cluster nodes...");
			result = partition.callMethodOnCluster(
					"EnergyComponents:name=ECSchedulerService", methodName,
					argValues, argTypes, false);
			log.info("...done, got respone from [" + result.size()
					+ "] nodes: \n" + result);
		} catch (InstanceNotFoundException e) {
			// if partition is not found, then call the local method directly
			// because we are not running in a clustering environment.
			result = new ArrayList();
			Method method = ECSchedulerService.class.getMethod(methodName,
					argTypes);
			if (method != null) {
				log.info("We're not in cluster environment, so call local method.");
				result.add(method.invoke(this, argValues));
			} else {
				log.error("No method found with signature: " + methodName + "("
						+ argValues + ", " + argValues + ")");
				throw new Exception("No method found with signature: "
						+ methodName + "(" + argValues + ", " + argValues + ")");
			}
		}
		**/

		return result;
	}

	/*
	 * protected Object executePinnedJob(String nodeID, Object event) throws
	 * JobExecutionException { if (hostID.startsWith(nodeID)) {
	 * log.info("About to execute pinned job on this node [" + hostID +"].");
	 * try { return
	 * callExtObject("com.ec.eccore.util.quartz.QuartzClusterHelper", null,
	 * null, "executeJobWrapper", new Class[] {Object.class}, new Object[]{
	 * event}); } catch (Exception e) {
	 * log.error("Failed to execute pinned job on node [" + hostID +"]: "
	 * +e.getMessage()); throw new JobExecutionException(e); } } else {
	 * log.info("Node [" + hostID
	 * +"] ignoring job, it's not for me. It is for node [" + nodeID +"]."); }
	 * return null; }
	 */

	// ############ JOBS #######################################
	public List getLocalRunningJobs(String sysname) throws SchedulerException,
			MalformedObjectNameException, AttributeNotFoundException,
			InstanceNotFoundException, MBeanException, ReflectionException {
		List list = new ArrayList();
		Scheduler sched = SchedulerRepository.getInstance().lookup(sysname);

		if (sched != null) {
			for (Iterator i = sched.getCurrentlyExecutingJobs().iterator(); i
					.hasNext();) {
				Map info = new HashMap();
				JobExecutionContext jobCtx = (JobExecutionContext) i.next();
				Job job = jobCtx.getJobInstance();
				info.put("hashCode", ""
						+ (job != null ? "" + job.hashCode() : null));
				info.put("name", jobCtx.getJobDetail().getKey().getName());
				info.put("group", jobCtx.getJobDetail().getKey().getGroup());
				info.put("firetime", ""
						+ jobCtx.getScheduledFireTime().getTime());
				info.put("node", hostID + "." + sysname);
				list.add(info);
			}
		}
		log.debug("Here are the results from the norwegian jury: \n" + list);
		return list;
	}

	public List getRunningJobs(String fullName) throws Exception {
		log.info("getRunningJobs()");
		return callClusterMethod("getLocalRunningJobs",
				new Class[] { String.class }, new Object[] { fullName });
	}

	// ############ RUNNING #######################################
	public String setLocalStateRunning(String sysname) {

		Scheduler sched = SchedulerRepository.getInstance().lookup(sysname);
		Properties prop = getQuartzProperties(sysname);

		if (sched != null && prop != null) {
			try {
				// Store the new state
				prop.setProperty(PROP_KEY_STARTUP_STATE, SERVER_STATE_RUNNING);
				// Set the state
				sched.start();
				return hostID + "." + sysname + "=" + SERVER_STATE_RUNNING;
			} catch (SchedulerException e) {
				log.info("Unable to set [" + SERVER_STATE_RUNNING + "] on ["
						+ sysname + "]: " + e.getMessage());
			}
		} else {
			log.info("Unable to set [" + SERVER_STATE_RUNNING + "] on server ["
					+ sysname + "], it's not alive.");
		}
		return hostID + "." + sysname + "=" + SERVER_STATE_SHUTDOWN;
	}

	public String setLocalStateRunning(String node, String sysname) {
		if (!hostID.equalsIgnoreCase(node)) {
			log.debug("Node [" + node + "] != [" + hostID
					+ "] => ignoring setLocalStateStandBy(" + node + ", "
					+ sysname + ")");
			return null;
		}
		return this.setLocalStateRunning(sysname);
	}

	public List setStateRunningOnAllNodes(String sysname) throws Exception {
		if (sysname.indexOf(".") >= 0) {
			log.warn("You need to supply a Sysname and not a FullName to this method.");
			return null;
		}
		return callClusterMethod("setLocalStateRunning",
				new Class[] { String.class }, new Object[] { sysname });
	}

	public String setStateRunning(String fullname) throws Exception {
		String name = this.getSysnameFromFullName(fullname);
		String node = this.getNodeNameFromFullName(fullname);
		log.debug("fullname: " + fullname + " sysname: " + name + " node: "
				+ node);

		if (name == null || node == null) {
			log.warn("A fullname is required for this method, not just a sysname.");
			return null;
		}

		List result = callClusterMethod("setLocalStateRunning", new Class[] {
				String.class, String.class }, new Object[] { node, name });
		for (Iterator iter = result.iterator(); iter.hasNext();) {
			String state = (String) iter.next();
			if (state != null) {
				return state;
			}
		}
		return null;
	}

	// ############ STANDBY #######################################
	public String setLocalStateStandBy(String sysname) {

		Scheduler sched = SchedulerRepository.getInstance().lookup(sysname);
		Properties prop = getQuartzProperties(sysname);

		if (sched != null && prop != null) {
			try {
				// Store the new state
				prop.setProperty(PROP_KEY_STARTUP_STATE, SERVER_STATE_STANDBY);
				// Set the state
				sched.standby();
				return hostID + "." + sysname + "=" + SERVER_STATE_STANDBY;
			} catch (SchedulerException e) {
				log.info("Unable to set [" + SERVER_STATE_STANDBY + "] on ["
						+ sysname + "]: " + e.getMessage());
			}
		} else {
			log.info("Unable to set [" + SERVER_STATE_STANDBY + "] on server ["
					+ sysname + "], it's not alive.");
		}
		return hostID + "." + sysname + "=" + SERVER_STATE_SHUTDOWN;
	}

	public String setLocalStateStandBy(String node, String sysname) {
		if (!hostID.equalsIgnoreCase(node)) {
			log.debug("Node [" + node + "] != [" + hostID
					+ "] => ignoring setLocalStateStandBy(" + node + ", "
					+ sysname + ")");
			return null;
		}
		return this.setLocalStateStandBy(sysname);
	}

	public List setStateStandByOnAllNodes(String sysname) throws Exception {
		if (sysname.indexOf(".") >= 0) {
			log.warn("You need to supply a Sysname and not a FullName to this method.");
			return null;
		}
		return callClusterMethod("setLocalStateStandBy",
				new Class[] { String.class }, new Object[] { sysname });
	}

	public String setStateStandBy(String fullname) throws Exception {
		String name = this.getSysnameFromFullName(fullname);
		String node = this.getNodeNameFromFullName(fullname);
		log.debug("fullname: " + fullname + " sysname: " + name + " node: "
				+ node);

		if (name == null || node == null) {
			log.warn("A fullname is required for this method, not just a sysname.");
			return null;
		}

		List result = callClusterMethod("setLocalStateStandBy", new Class[] {
				String.class, String.class }, new Object[] { node, name });
		for (Iterator iter = result.iterator(); iter.hasNext();) {
			String state = (String) iter.next();
			if (state != null) {
				return state;
			}
		}
		return null;
	}

	// ############ INTERRUPT ##################################
	public boolean interruptLocalJobs(String sysname, String groupName,
			String jobName, String jobInstanceIdHashCode) {
		boolean interrupted = false;
		Scheduler quartzServer = SchedulerRepository.getInstance().lookup(
				sysname);
		if (quartzServer == null) {
			log.warn("Quartz Server: [" + sysname
					+ "] not found. Can't interrupt job [" + groupName + "."
					+ jobName + "].");
			return interrupted;
		}

		try {
			if (jobInstanceIdHashCode == null
					|| jobInstanceIdHashCode.length() <= 0) {
				JobKey jobKey = new JobKey(jobName, groupName);
				// Interrupt ALL running jobs with 'groupName.jobName'
				interrupted = quartzServer.interrupt(jobKey);
				log.info("Interrupting all jobs named [" + groupName + "."
						+ jobName + "] on Quartz Server [" + sysname
						+ "] returned: " + interrupted);
			} else {
				log.info("Trying to find Job with Hash ["
						+ jobInstanceIdHashCode + "]...");
				for (Iterator i = quartzServer.getCurrentlyExecutingJobs()
						.iterator(); i.hasNext();) {
					JobExecutionContext jobCnx = (JobExecutionContext) i.next();
					int jobHash = jobCnx.getJobInstance().hashCode();
					log.debug("Matching: " + jobInstanceIdHashCode + " <--> "
							+ jobHash);
					if (jobInstanceIdHashCode.equalsIgnoreCase("" + jobHash)) {
						// Interrupt only the given instance of Job
						Job job = jobCnx.getJobInstance();
						if (job instanceof InterruptableJob) {
							InterruptableJob intrJob = (InterruptableJob) job;
							intrJob.interrupt();
							interrupted = true;
							log.info("Interrupting job named [" + groupName
									+ "." + jobName + "] with Hash [" + jobHash
									+ "] on Quartz Server [" + sysname
									+ "] returned: " + interrupted);
						} else {
							log.warn("Jobs named ["
									+ groupName
									+ "."
									+ jobName
									+ "] with Hash ["
									+ jobHash
									+ "] on Quartz Server ["
									+ sysname
									+ "] does not implement the InterruptableJob interface.");
						}
						break;
					}
				}
			}
		} catch (UnableToInterruptJobException e) {
			log.error("Interrupting jobs [" + groupName + "." + jobName
					+ "] on Quartz Server: [" + sysname + "] failed: "
					+ e.getMessage());
		} catch (SchedulerException e) {
			log.error("Interrupting jobs [" + groupName + "." + jobName
					+ "] on Quartz Server: [" + sysname + "] failed: "
					+ e.getMessage());
		}
		return interrupted;
	}

	public List interruptJobs(String sysname, String groupName, String jobName)
			throws Exception {
		return callClusterMethod("interruptLocalJobs", new Class[] {
				String.class, String.class, String.class, String.class },
				new Object[] { sysname, groupName, jobName, null });
	}

	public List interruptJob(String sysname, String jobInstanceIdHashCode)
			throws Exception {
		return callClusterMethod("interruptLocalJobs", new Class[] {
				String.class, String.class, String.class, String.class },
				new Object[] { sysname, null, null, jobInstanceIdHashCode });
	}

	// ############ INFO #######################################
	public List getInfoFromAllNodes(String sysname) throws Exception {
		return callClusterMethod("getLocalInfo", new Class[] { String.class },
				new Object[] { sysname });
	}

	public Map getLocalInfo(String sysname) {
		log.debug("getLocalInfo called");
		if (sysname.indexOf(".") >= 0) {
			log.warn("You need to supply a Sysname and not a FullName to this method.");
			return null;
		}

		Map map = new HashMap();
		Scheduler sched = SchedulerRepository.getInstance().lookup(sysname);
		String state = SERVER_STATE_SHUTDOWN;
		int usedPoolSize = 0;
		long runningSince = 0;
		int totalExecutedCount = 0;
		int poolSize = 0;
		try {
			if (sched != null && !sched.isShutdown()) {
				SchedulerMetaData meta = sched.getMetaData();
				poolSize = meta.getThreadPoolSize();
				totalExecutedCount = meta.getNumberOfJobsExecuted();
				Date runSinceDate = meta.getRunningSince();
				if (runSinceDate != null) {
					runningSince = runSinceDate.getTime();
				}
				usedPoolSize = sched.getCurrentlyExecutingJobs().size();
			}
			if (sched != null && !sched.isInStandbyMode()
					&& !sched.isShutdown()) {
				state = SERVER_STATE_RUNNING;
			} else if (sched != null && sched.isInStandbyMode()) {
				state = SERVER_STATE_STANDBY;
			}
		} catch (SchedulerException e) {
			state = SERVER_STATE_SHUTDOWN;
			log.debug("Caught an exception: " + e.getMessage(), e);
		}

		int currentOffset = 0;
		try {
			Integer iobj = (Integer) this.callExtObject(
					"com.ec.eccore.util.date.DateHelper", new Class[] {},
					new Object[] {}, "getCurrentTimeZoneDiff",
					new Class[] { String.class }, new Object[] { sysname });
			currentOffset = iobj.intValue();
		} catch (Exception e) {
			log.debug("Couldn't call getCurrentTimeZoneDiff");
		}
		Properties prop = getQuartzProperties(sysname);
		String startupState = prop.getProperty(PROP_KEY_INITIAL_STARTUP_STATE)
				.trim();
		log.debug("Got the startup state : " + startupState);
		map.put("node", hostID + "." + sysname);
		map.put("startstate", startupState);
		map.put("currentstate", "" + state);
		map.put("systimeoffset", "" + currentOffset);
		map.put("maxthreads", "" + poolSize);
		map.put("execcount", "" + totalExecutedCount);
		map.put("starttime", "" + runningSince);
		map.put("freethreads", "" + (poolSize - usedPoolSize));
		log.debug("The Local Map is : " + map);

		return map;
	}

	// ############ STATE #######################################
	public String getLocalState(String sysname) {
		if (sysname.indexOf(".") >= 0) {
			log.warn("You need to supply a Sysname and not a FullName to this method.");
			return null;
		}

		Scheduler sched = SchedulerRepository.getInstance().lookup(sysname);
		String state = SERVER_STATE_SHUTDOWN;

		try {
			if (sched != null && !sched.isInStandbyMode()
					&& !sched.isShutdown()) {
				state = SERVER_STATE_RUNNING;
			} else if (sched != null && sched.isInStandbyMode()) {
				state = SERVER_STATE_STANDBY;
			}
		} catch (SchedulerException e) {
			state = SERVER_STATE_SHUTDOWN;
		}
		return hostID + "." + sysname + "=" + state;
	}

	public String getLocalState(String node, String sysname) {
		if (!hostID.equalsIgnoreCase(node)) {
			log.debug("Node [" + node + "] != [" + hostID
					+ "] => ignoring getLocalState(" + node + ", " + sysname
					+ ")");
			return null;
		}
		return this.getLocalState(sysname);
	}

	public String getState(String fullName) throws Exception {
		String name = this.getSysnameFromFullName(fullName);
		String node = this.getNodeNameFromFullName(fullName);
		log.debug("fullname: " + fullName + " sysname: " + name + " node: "
				+ node);

		List result = callClusterMethod("getLocalState", new Class[] {
				String.class, String.class }, new Object[] { node, name });
		for (Iterator iter = result.iterator(); iter.hasNext();) {
			String state = (String) iter.next();
			if (state != null) {
				return state;
			}
		}
		return null;
	}

	public List getStateFromAllNodes(String sysname) throws Exception {
		return callClusterMethod("getLocalState", new Class[] { String.class },
				new Object[] { sysname });
	}

	// ############ NODE #######################################
	public String getLocalNodeName() {
		return hostID;
	}

	public List getNodeNames() throws Exception {
		return callClusterMethod("getLocalNodeName", new Class[] {},
				new Object[] {});
	}

	public String getNodeNameFromFullName(String fullName) {
		if (fullName == null) {
			return null;
		}
		int pos = fullName.lastIndexOf(".");
		return (pos >= 0 ? fullName.substring(0, pos) : null);
	}

	public String getSysnameFromFullName(String fullName) {
		if (fullName == null) {
			return null;
		}
		int pos = fullName.lastIndexOf(".");
		return (pos >= 0 ? fullName.substring(pos + 1) : null);
	}

	// ############ SYSTEM #######################################
	public Collection getLocalSystemNames() {
		try {
			return this.getEcSystems();
		} catch (Exception e) {
			log.info("Unable to retrive System Names: " + e.getMessage());
		}
		return null;
	}

	public Collection getSystemNames() throws Exception {
		Set s = new HashSet();
		List list = callClusterMethod("getLocalSystemNames", new Class[] {},
				new Object[] {});
		for (Iterator i = list.iterator(); i.hasNext();) {
			s.addAll((Collection) i.next());
		}
		return s;
	}

	// ############ DUMP CONFIG XML #######################################
	public void dumpToPath(String path) throws Exception {
		callExtObject("com.ec.frmw.scheduler.model.ScheduleSerializer",
				new Class[] {}, new Object[] {}, "dumpToPath",
				new Class[] { String.class }, new Object[] { path });
	}

	public void dumpToPath(String path, String name, String group)
			throws Exception {
		callExtObject("com.ec.frmw.scheduler.model.ScheduleSerializer",
				new Class[] {}, new Object[] {}, "dumpToPath", new Class[] {
						String.class, String.class, String.class },
				new Object[] { path, name, group });
	}

	public void loadFromPath(String path) throws Exception {
		callExtObject("com.ec.frmw.scheduler.model.ScheduleSerializer",
				new Class[] {}, new Object[] {}, "loadFromPath",
				new Class[] { String.class }, new Object[] { path });
	}

	// ###################################################
	protected Object callExtObject(String className, Class constructorTypes[],
			Object constructorArgList[], String methodName,
			Class methodTypes[], Object methodArgs[])
			throws ClassNotFoundException, SecurityException,
			NoSuchMethodException, IllegalArgumentException,
			InstantiationException, IllegalAccessException,
			InvocationTargetException, ExceptionInInitializerError {
		// Load the class
		Class cls = Class.forName(className);

		Constructor ct = null;
		Object object = null;

		// Try to find a constructor of this type on the class?
		ct = cls.getConstructor(constructorTypes);

		// Try to instantiate the class
		object = ct.newInstance(constructorArgList);

		// Try to find a method of this type on the class?
		Method method = cls.getMethod(methodName, methodTypes);

		// Execute the method!
		Object ret = method.invoke(object, methodArgs);
		return ret;
	}

	private void connect(String hostname, String port) throws IOException{
//		Integer portInteger = Integer.valueOf(port);
//		Hashtable h = new Hashtable();
//		JMXServiceURL address = new JMXServiceURL(
//				"service:jmx:rmi:///jndi/rmi://" + hostname + ":" + port
//						+ "/jmxrmi");
//		connector = JMXConnectorFactory.connect(address, null);
//		connection = connector.getMBeanServerConnection();
		connection = MBeanServerFactory.newMBeanServer();
//		connection = ManagementFactory.getPlatformMBeanServer();
		log.info("GOT THE MBeanServerConnection---SUCCESSFULLY");
	}

}
